import { createSelector } from 'reselect';
import dedupe from 'promise-dedupe';

export const configs = {
  REDUCER_NAMESPACE: '@@dequest',
};

class Dequest {
  constructor(requestId, request, { overrideTransform, apiEntry } = {}) {
    this.requestId = requestId;
    this.request = request;
    this.options = { overrideTransform, apiEntry };
  }
}

export class HandledRequest {
  constructor({ type, args } = {}) {
    this.type = type;
    this.args = args;
  }
}

export const get = (...args) => new HandledRequest({ type: 'get', args });
export const post = (...args) => new HandledRequest({ type: 'post', args });
export const put = (...args) => new HandledRequest({ type: 'put', args });
export const Delete = (...args) => new HandledRequest({ type: 'delete', args }); // damn, delete is reserved
export const patch = (...args) => new HandledRequest({ type: 'patch', args });

export const makeRequest = (
  requestId,
  request,
  { overrideTransform = false, apiEntry } = {},
) => {
  return new Dequest(requestId, request, { overrideTransform, apiEntry });
};

export const invalidateRequest = requestId => ({
  type: '@@DEQUEST/INVALIDATE',
  requestId,
});

export const $selectResponse = createSelector(
  state => state,
  state => state[configs.REDUCER_NAMESPACE],
);

export const defaultReducer = {
  [configs.REDUCER_NAMESPACE]: (state = {}, action) => {
    switch (action.type) {
      case '@@DEQUEST/SEND':
        return {
          ...state,
          [action.requestId]: {
            isLoading: true,
            requestAt: action.requestAt,
            responseAt: null,
          },
        };
      case '@@DEQUEST/RECEIVE':
        return {
          ...state,
          [action.requestId]: {
            ...state[action.requestId],
            ...action.data,
            isLoading: false,
            isUpdating: false,
            responseAt: action.responseAt,
          },
        };
      case '@@DEQUEST/UPDATE':
        return {
          ...state,
          [action.requestId]: {
            ...state[action.requestId],
            isUpdating: true,
            exception: null,
            requestAt: action.requestAt,
            responseAt: null,
          },
        };
      case '@@DEQUEST/INVALIDATE':
        return {
          ...state,
          [action.requestId]: undefined,
        };
      default:
        return state;
    }
  },
};

const resolveRequest = async (api, action) => {
  if (action.request instanceof HandledRequest) {
    if (!api) throw new Error('Dequest: custom api is not defined');
    const type = action.request.type;
    const method = api[type];
    if (!method) {
      throw new Error(`Dequest: method ${type} not registered in custom api`);
    }
    return method(action);
  }
  return action.request;
};

const noOpTransformer = _ => Promise.resolve(_);

const resolveApiEntry = ({ api, apis, entry }) => {
  if (!api && !apis) {
    return undefined;
  }

  if (api && apis) {
    throw new Error('`api` and `apis` could not be both defined');
  }

  if (api) return api;

  if (!entry) {
    throw new Error(
      'Dequest: `apis` defined in middleware but `apiEntry` not found in action',
    );
  }

  if (!apis[entry]) {
    throw new Error(`Entry ${entry} not found in 'apis'`);
  }

  return apis[entry];
};

export const createMiddleware = (
  { requestTransformer = noOpTransformer, api, apis } = {},
) => store => next => action => {
  if (!(action instanceof Dequest)) {
    return next(action);
  }

  const requestAt = Date.now();
  const { $isFinished } = createResponseSelectorsByKey(action.requestId);
  if ($isFinished(store.getState())) {
    store.dispatch({
      type: '@@DEQUEST/UPDATE',
      requestId: action.requestId,
      requestAt,
    });
  } else {
    store.dispatch({
      type: '@@DEQUEST/SEND',
      requestId: action.requestId,
      requestAt,
    });
  }

  const transform = action.options.overrideTransform
    ? action.options.overrideTransform
    : requestTransformer;

  const apiEntry = resolveApiEntry({
    api,
    apis,
    entry: action.options.apiEntry,
  });

  return transform(dedupe(action.requestId, resolveRequest(apiEntry, action)))
    .then(r =>
      store.dispatch({
        type: '@@DEQUEST/RECEIVE',
        requestId: action.requestId,
        data: r,
        responseAt: Date.now(),
      }),
    )
    .catch(error => {
      console.error(error);
      throw new Error(
        'Error: promise rejected. ' +
          'Dequest requires request promise to always resolves!',
      );
    });
};

export const defaultMiddleware = createMiddleware();

export const isStartedResolver = response => !!response;
export const bodyResolver = response => response && response.jsonBody;
export const isLoadingResolver = (isStarted, response) =>
  isStarted && response.isLoading === true;
export const isUpdatingResolver = response =>
  response && response.isUpdating === true;
export const isSuccessResolver = response => response && response.ok;
export const isErrorResolver = (isLoading, response) =>
  !isLoading && response && (response.ok === false || !!response.exception);
export const errorResolver = (response, isError) =>
  response && isError && (response.jsonBody || response.exception);
export const exceptionResolver = response => response && response.exception;
export const isFinishedResolver = response =>
  response && response.ok !== undefined;
export const requestAtResolver = response => response && response.requestAt;
export const responseAtResolver = response => response && response.responseAt;
export const requestDurationResolver = (requestAt, responseAt) =>
  !!requestAt && !!responseAt ? responseAt - requestAt : null;
export const statusCodeResolver = response => response && response.status;

export const createResponseSelectors = $response => {
  const $isStarted = createSelector($response, isStartedResolver);
  const $isLoading = createSelector($isStarted, $response, isLoadingResolver);
  const $isUpdating = createSelector($response, isUpdatingResolver);
  const $isSuccess = createSelector($response, isSuccessResolver);
  const $isError = createSelector($isLoading, $response, isErrorResolver);
  const $body = createSelector($response, bodyResolver);
  const $error = createSelector($response, $isError, errorResolver);
  const $exception = createSelector($response, exceptionResolver);
  const $isFinished = createSelector($response, isFinishedResolver);
  const $requestAt = createSelector($response, requestAtResolver);
  const $responseAt = createSelector($response, responseAtResolver);
  const $requestDuration = createSelector(
    $requestAt,
    $responseAt,
    requestDurationResolver,
  );
  const $statusCode = createSelector($response, statusCodeResolver);

  return {
    $isStarted,
    $isUpdating,
    $isFinished,
    $isLoading,
    $isSuccess,
    $body,
    $isError,
    $error,
    $exception,
    $requestAt,
    $responseAt,
    $requestDuration,
    $statusCode,
  };
};

export const createResponseSelectorsByKey = key => {
  const $response = createSelector(
    $selectResponse,
    selectResponse => selectResponse[key],
  );

  const selectors = {
    $response,
    ...createResponseSelectors($response),
  };

  return selectors;
};

export const isDequestReceiveAction = object =>
  typeof object === 'object' &&
  object.type &&
  object.type === '@@DEQUEST/RECEIVE';

export default {
  reducer: defaultReducer,
  middleware: defaultMiddleware,
  createMiddleware: createMiddleware,
};
